@@ -9,7 +9,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <fcntl.h>
-#include <sys/poll.h>
+#include <poll.h>
 
 #include "liburing.h"
 
